"use strict"

const marginSuffix = "[M]"
const isoSuffix = "[I]"
const ctSuffix = "[C]"

const fieldSymbol = ["symbol", "instrument", "pair", "product_id", "marketSymbol", "instrument_name", "instId"]
const fieldAvgPrice = ["avgPrice", "avgPx", "avg_execution_price", "avgFillPrice"]
const fieldPrice = ["price", ...fieldAvgPrice, "limitPrice", "orderPrice", "px"]
const fieldTP = ["take_profit", "takeProfit", "tpTriggerPx", "tpLimitPrice"]
const fieldSL = ["stop_loss", "stopLoss", "slTriggerPx", "slLimitPrice"]
const fieldStop = ["stopPx", "stop_px", "stopprice", "activationPrice", "triggerPrice", "trigger_price", ...fieldTP, ...fieldSL, "trailing_stop"]
const fieldSide = ["side", "direction"]
const fieldHedge = ["positionSide", "posSide", "position_idx", "positionIdx"]
const fieldQuantityExec = ["executedQty", "cum_exec_qty", "cumExecQty", "fillQuantity", "filled_amount", "filledSize", "filled_size", "filled", "fillSz"]
const fieldQuantityLeft = ["leavesQty", "leaves_qty", "leavesQty", "remainingSize", "remaining_amount"/*, "amount"*/, "unfilledSize"]
const fieldQuantity = ["orderQty", "quantity", "units", "qty", "size", "original_amount", "startingAmount", "origQty", "volume", "sz", "amount", "orderQtyRq", "sizeRq"]
const fieldType = ["ordType", "type", "order_type", "ordertype", "orderType"]
const fieldTIF = ["timeInForce", "time_in_force"]
const fieldStatus = ["ordStatus", "status", "order_status", "orderStatus"]
const fieldId = ["orderID", "order_id", "ordId", "id", "txid", "algoId", "comment"]
const fieldCustom = ["algoClOrdId", "clOrdId", "clOrdID"/*, "uuid"*/, "newClientOrderId", "clientOrderId", "order_link_id", "orderLinkId", "clientOid", "client_oid", "label", "clientId", "cliOrdID"]
const fieldAll = [...fieldSymbol, ...fieldPrice, ...fieldStop, ...fieldSide, ...fieldHedge, ...fieldQuantity, ...fieldQuantityExec, ...fieldQuantityLeft, ...fieldType, ...fieldTIF, ...fieldStatus, ...fieldId, ...fieldCustom]

const fieldPosSide = [...fieldSide, "type", "positionSide", "posSide"]
const fieldPosHedge = fieldHedge
const fieldPosQuantity = ["quantity", /*"size_currency",$$$*/ "size", "volume", "amount", "available", /*"availPos",*/ "pos"]
const fieldPosPrice = ["avgEntryPrice", "recentAverageOpenPrice", "entryPrice", "entry_price", "average_price", "averagePrice", "avgPrice", "openPrice", "price_avg", "basePrice", "price", "avgPx", "base"]
const fieldPosPriceEven = ["breakEvenPrice", "recentBreakEvenPrice"]
const fieldPosPriceMark = ["markPrice", "mark_price", "mark", "last"]
const fieldPosPriceLiq = ["liquidationPrice", "liq_price", "liqPrice", "estimatedLiquidationPrice", "estimated_liquidation_price", "liqPx"]
const fieldPosProfit = ["unrealisedPnl", "unRealizedProfit", "unrealised_pnl", "unrealizedPL", "recentPnl", "unrealizedPnl", "floating_profit_loss", "profit", "net", "pl", "upl"]
const fieldPosDate = ["createdAt", "openingTimestamp", "updated_at", "updateTime", "updatedTime", "create_date", "createTime", "createdTime", "openTime", "cTime", "time", "transactTimeNs"]

const errOrderID = "Order ID check failed!"
const errSlippage = "Price difference too high (slippage)"
const errMinSpread = "Bid/Ask difference too low (min spread)"
const errMaxSpread = "Bid/Ask difference too high (max spread)"

class Exchange {
	constructor(meta) {
		!meta.nonceName && (meta.nonceName = meta.aliases[0] + "_NONCE")
		!meta.symbolCacheName && (meta.symbolCacheName = meta.aliases[0] + "_SYMBOLS")
		!meta.offsetName && (meta.offsetName = meta.aliases[0] + "_OFFSET")
		this.meta = meta
		this.cmd = {logLevel: 15}
	}

	hasMargin() {
		return this.meta.margin
	}
	hasSpot() {
		return this.meta.spot || !this.meta.margin
	}
	canListPos() {
		return this.meta.margin && this.meta.listPos !== false
	}
	getAlias(index = 0) {
		return this.meta.aliases && (index < 0 ? this.meta.aliases.last() : this.meta.aliases[index])
	}
	getAliases() {
		return this.meta.aliases || (() => {throw new ReferenceError("Exchange aliases are not configured!")})()
	}
	getFields() {
		return this.meta.fields || (() => {throw new ReferenceError("Exchange fields are not configured!")})()
	}
	getName() {
		return this.meta.name || (() => {throw new ReferenceError("Exchange name is not configured!")})()
	}
	getPatterns() {
		return this.meta.patterns || []
	}
	getPermissions() {
		return this.meta.permissions || (() => {throw new ReferenceError("Exchange permissions are not configured!")})()
	}
	getWebsite() {
		return this.meta.website || (() => {throw new ReferenceError("Exchange website is not configured!")})()
	}
	getDescription() {
		return this.meta.desc || (() => {throw new ReferenceError("Exchange description is not configured!")})()
	}
	getDefaultSymbol() {
		return this.meta.defaultSymbol || "BTC"
	}
	getTestSymbol() {
		return this.getAlias() + ':' + (this.meta.testSymbol || "XBTUSD")
	}
	getAccounts() {
		return Object.keys(syncStore.get('exchanges', this.getAlias()) || {})
	}
	getAccountRaw(acc) {
		return syncStore.get('exchanges', this.getAlias(), acc)
	}
	hasAccount(acc) {
		return this.getAccounts().includesNoCase(acc)
	}
	hasCredentials() {
		return this.getAccounts().length
	}
	getAccount() {
		return this.cmd.a
	}
	setAccount(acc) {
		this.cmd.a = acc
	}
	setEventID(id) {
		this.cmd.id = id
	}
	setLogLevel(level) {
		this.cmd.logLevel = level
	}
	debug(msg) {
		if (this.cmd.logLevel & 1) log.debug((this.cmd.id ? `[${this.cmd.id}] ` : '')+this.getName()+" "+msg)
	}
	info(msg) {
		if (this.cmd.logLevel >= 14) {
			log.info((this.cmd.id ? `[${this.cmd.id}] ` : '')+this.getName()+" "+msg)
			this.cmd.telebot && telebotSend(this.cmd.telebot, (this.cmd.id ? `[${this.cmd.id}] ` : '')+this.getName()+" "+msg)
		}
	}
	warn(msg) {
		if (this.cmd.logLevel & 4) {
			log.warn((this.cmd.id ? `[${this.cmd.id}] ` : '')+this.getName()+" "+msg)
			this.cmd.telebot && telebotSend(this.cmd.telebot, (this.cmd.id ? `[${this.cmd.id}] ` : '')+this.getName()+" "+msg)
		}
	}
	notice(msg) {
		if (this.cmd.logLevel & 8) {
			log.notice((this.cmd.id ? `[${this.cmd.id}] ` : '')+this.getName()+" "+msg)
			this.cmd.telebot && telebotSend(this.cmd.telebot, (this.cmd.id ? `[${this.cmd.id}] ` : '')+this.getName()+" "+msg)
		}
	}

	static cleanSym = /[-_/\\:]/g

	sym(name) {
		return name.replace(Exchange.cleanSym, '')
	}

	getSubscriptions(key) {
		this.meta.subscriptions || (() => {throw new ReferenceError("Exchange subscriptions are not configured!")})()
		if (key) {
			return this.meta.subscriptions[key] || (() => {throw new ReferenceError("Invalid subscription key "+key)})()
		}
		return this.meta.subscriptions
	}

	getCredentials(field = 'public') {
		const acc = this.getAccount() || (() => {throw new Error("No account provided!")})()
		this.hasAccess()
		const accs = syncStore.get('exchanges', this.getAlias())
		const cred = (accs || {}).getProp(acc)
		if (!accs || !cred || !cred[field]) {
			const names = Object.keys(accs || {})
			names.length || (() => {throw new ReferenceError(`No accounts configured for ${this.getName()} – please check "Options / Accounts"!`)})()
			throw new ReferenceError(`Non-existent account ${acc} specified for ${this.getName()} (available: ${names.join(', ')}) – please check "Options / Accounts"!`)
		}
		const p0 = "Yzk5ZjZmYzliOWQ0ODA"
		const p1 = "Y2E4ZGUxYjRhN2N"
		const p2 = localStore.gets(_('cGVybXMucGVybWlzc2lvbnM'))
		const p3 = "hODdjNGIx"
		if ((this.getSubscriptions().active.length || this.getSubscriptions().inactive.length) && Object.values(p2||0).indexOf(_(p1 + p3 + p0)) < 0) {
			throw new ReferenceError(this.getName() + _('OiBVbmZvcnR1bmF0ZWx5IHRoaXMgaXMgb25seSBzdXBwb3J0ZWQgaW4gUHJvZml0VmlldyBQUk8h'))
		}
		return cred
	}

	hasPermission() {
		const exPerms = this.getPermissions()
		const legacyPerms = getLegacyPermissions(exPerms)
		return perms.hasAll(exPerms) || perms.hasAll(legacyPerms[0]) || perms.hasAll(legacyPerms[1])
	}

	hasSubscription() {
		const subs = this.getSubscriptions()
		if (!subs.active.length && 
			(this.meta.name.includes(_("VGVzdG5ldA")) || this.meta.name.includes(_("U2FuZGJveA")) || this.meta.name.includes(_("RGVtbw")))) {
			return subs.inactive
		}
		return perms.hasAll(window.ACCESS) && perms.hasAny({permissions: subs.active})
	}

	hasAccess() {
		this.hasPermission() || (() => {throw new ReferenceError(this.getName() + _('OiBQbGVhc2UgZ3JhbnQgbmVjZXNzYXJ5IHBlcm1pc3Npb25zIHVuZGVyICJTZXR1cDogUGVybWlzc2lvbnMiIQ'))})()
		this.hasSubscription() || (() => {throw new ReferenceError(this.getName() + _('OiBVbmZvcnR1bmF0ZWx5IHRoaXMgaXMgb25seSBzdXBwb3J0ZWQgaW4gUHJvZml0VmlldyBQUk8h'))})()
	}

	static illegalAcc = /[ \t=,'"#:\[\]()]/g

	setCredentials(acc, input) {
		if (!acc.length || !acc.trim()) {
			throw new SyntaxError("Account name was not provided!")
		}
		if (Exchange.illegalAcc.test(acc)) {
			throw new SyntaxError("Account name contains reserved or illegal characters!")
		}
		let data = {account: acc}
		for (const field in this.meta.fields) {
			let val = input[field]
			if (typeof val === 'number') {
				data[field] = val
			} else if (val) {
				if (typeof val !== 'string') {
					throw new TypeError(`Field was not a string! (${typeof val})`)
				}
				val = val.replace(/[^\x00-\x7F]/g, '').trim()
				if (!val.length && !this.meta.fields[field].optional) {
					throw new SyntaxError(`Field '${this.meta.fields[field].label}' was not provided!`)
				}
				data[field] = val
			} else if (!this.meta.fields[field].optional) {
				throw new SyntaxError(`Field '${this.meta.fields[field].label}' was not provided!`)
			}
		}
		return data
	}

	*executeCommand(cmd, id, ohlc = {}, logLevel = 15, telebot) {
		this.setAccount(cmd.a)
		this.setEventID(id)
		this.setLogLevel(logLevel)
		telebot && (this.cmd.telebot = telebot)
		this.loadOrderData(cmd, ohlc)

		if ((cmd.ch || cmd.c) === 'order') {
			return this.saveOrderData(cmd, ohlc, yield* this.ordersCancel(cmd))
		} else if ((cmd.ch || cmd.c) === 'position') {
			if (cmd.c && cmd.cid) {
				const pos = (yield* this.getPositions(cmd.a, cmd.s, 1))[cmd.a].list
				if (!pos || !pos.length) {
					this.info("no positions open!")
					return false
				}
				const ord = (yield* this.getOrders(cmd.a, cmd.s, logLevel, cmd.cid, true))[cmd.a]
				if (!ord || !ord.length) {
					throw new Error(errOrderID+` (cid=${cmd.cid} specified)`)
				}
				return this.saveOrderData(cmd, ohlc, yield* this.positionsCloseAll(cmd.clone({c: 'position', ch: undefined}), ohlc, pos))
			}
			return this.saveOrderData(cmd, ohlc, yield* this.positionsCloseAll(cmd, ohlc))
		} else if (cmd.ch || cmd.c) {
			throw new SyntaxError(`Invalid cancel/close/check option "${cmd.ch||cmd.c}" (${cmd.ch?'ch':'c'}=) specified!`)
		}

		if (cmd.tb) {
			if (!this.transfer) throw new SyntaxError(`Command "transfer balance" (tb=) not supported on this exchange!`)
			return yield* this.transfer(cmd, ohlc)
		} else if (cmd.loan) {
			if (!this.loan) throw new SyntaxError(`Command "loan margin" (loan=) not supported on this exchange!`)
			return yield* this.loan(cmd, ohlc)
		} else if (cmd.repay) {
			if (!this.repay) throw new SyntaxError(`Command "repay margin" (repay=) not supported on this exchange!`)
			return yield* this.repay(cmd, ohlc)
		} else if (cmd.convert) {
			if (!this.convert) throw new SyntaxError(`Command "convert balance" (convert=) not supported on this exchange!`)
			return yield* this.convert(cmd, ohlc)
		} else if (cmd.withdraw) {
			if (!this.withdraw) throw new SyntaxError(`Command "withdraw balance" (withdraw=) not supported on this exchange!`)
			return yield* this.withdraw(cmd, ohlc)
		}

		if (cmd.b || cmd.ub || cmd.up) {
			if (['limit', 'market', 'fok', 'ioc', 'post', 'day', 'settle'].includes(cmd.t)) {
				if (['pos', 'even', 'liq'].includes(cmd.pr) || cmd.y === 'possize') {
					const pos = (yield* this.getPositions(cmd.a, cmd.s, -1/*, cmd.b ??? */))[cmd.a].list
					this.setLogLevel(logLevel)
					if (!pos || !pos.length) {
						this.warn("position based price reference (pr=pos/even/liq) or trade size (y=pos) specified but no position open!")
						if (cmd.nr) return false
					} else {
						ohlc.possize = 0
						for (const entry of pos) {
							ohlc.pos = entry._entry
							ohlc.even = entry._even
							ohlc.liq = entry._liq
							ohlc.possize += entry._size
						}
					}
				}
				return this.saveOrderData(cmd, ohlc, yield* this.trade(cmd, ohlc))
			}
			throw new SyntaxError(`Invalid order type "${cmd.t}" (t=) specified!`)
		}
		!cmd.lock && this.warn("Empty command executed!")
	}

	static formatCcy(name) {
		return name.toUpperCase().replace(ctSuffix+marginSuffix, '<sup>CT</sup>').replace(isoSuffix+marginSuffix, '<sup>ISO</sup>').replace(marginSuffix, '<sup>M</sup>')
	}

	updateBalance(balance, symbol, isMargin, noInfo) {
		if (!balance) return
		symbol = symbol || this.getDefaultSymbol()
		if (balance.balance || balance.Balance || balance.total || balance.Total) {
			balance = {[symbol]: balance}
		}
		if (!noInfo) this.notice(`available ${isMargin?'margin ':''}balance(s): `+stringify(balance))

		if (isMargin) {
			let margin = {}
			for (const asset in balance) {
				margin[typeof balance[asset] === 'object' && !asset.endsWith(marginSuffix) ? asset+marginSuffix : asset] = balance[asset]
			}
			balance = margin
			if (!symbol.endsWith(marginSuffix)) {
				symbol += marginSuffix
			}
		}
		balance.lastSymbol = symbol
		balance.updated = new Date().toISOString()
		localStore.set('balances', this.getAlias(), this.getAccount().toUpperCase(), balance)
	}

	getBalanceRaw(acc) {
		return localStore.get('balances', this.getAlias(), (acc || this.getAccount() || '*').toUpperCase()) || {}
	}
	setBalanceRaw(key, val, acc) {
		localStore.set('balances', this.getAlias(), (acc || this.getAccount() || '*').toUpperCase(), key, val)
	}

	getBalance(symbol, isMargin, noInfo) {
		const acc = this.getAccount()
		const balances = this.getBalanceRaw()
		symbol = symbol || (balances && balances.lastSymbol) || this.getDefaultSymbol()
		if (isMargin && !symbol.endsWith(marginSuffix)) {
			symbol += marginSuffix
		}
		if (!noInfo) if (!balances) {
			this.warn(`does not have any cached balance info yet for account ${acc}, querying current balance! (use ub=1 or non-bc somewhere before bc=1)`)
		} else if (!balances[symbol]) {
			this.warn(`does not have cached balance info yet for account ${acc} and symbol ${symbol}, querying current balance! (use ub=1 or non-bc somewhere before bc=1)`)
		} else {
			this.info(`using cached balance info for account ${acc} and symbol ${symbol}! (bc=1 specified)`)
		}
		return balances && balances[symbol]
	}

	updatePrice(symbol, ticker, isMargin) {
		if (isMargin) {
			symbol += typeof isMargin === "string" ? `[${isMargin}]` : marginSuffix
		}
		this.notice(`caching ticker for ${symbol}: `+stringify(ticker))
		ticker.updated = new Date().toISOString()
		localStore.set('tickers', this.getAlias(), symbol, ticker)
	}

	getPrice(symbol, isMargin, noInfo) {
		if (isMargin) {
			symbol += typeof isMargin === "string" ? `[${isMargin}]` : marginSuffix
		}
		const ex = localStore.get('tickers', this.getAlias())
		if (!noInfo) if (!ex) {
			this.warn("does not have any cached price data yet, using current price! (use up=1 somewhere before pc=1)")
		} else if (!ex[symbol]) {
			this.warn(`does not have cached price data for ${symbol} yet, using current price! (use up=1 somewhere before pc=1)`)
		} else {
			this.info(`using cached price data for ${symbol}! (pc=1 specified)`)
		}
		return ex && ex[symbol]
	}

	uniqueID(id, suffix1, suffix2) {
		let event = (this.cmd.id || '').toString()
		const subs = event.split('=>')
		const subnum = subs.length - (event.charCodeAt(0) >= 48 && event.charCodeAt(0) <= 57 ? 2 : 1)
		let line = ((subs.last().match(/:([\d:]+)/)||[])[1] || '').replace(':' , '-')
		event = (event.match(/^[\d\-]+/)||[])[0] || this.cmd.randomID || (this.cmd.randomID = (Date.now() % 100000000).toString())
		return (id ? `${id}-` : '') + event + (subnum > 0 ? `-${subnum}` : '') + (line ? `-${line}` : '') + (suffix1 ? `-${suffix1}` : '') + (suffix2 ? `-${suffix2}` : '')
	}

	uniqueIDA(len, id, suffix1, suffix2) {
		let event = (this.cmd.id || '').toString()
		const subs = event.split('=>')
		const subnum = subs.length - (event.charCodeAt(0) >= 48 && event.charCodeAt(0) <= 57 ? 2 : 1)
		let line = ((subs.last().match(/:([\d:]+)/)||[])[1] || '').replace(':' , 'R')
		event = (event.match(/^[\d\-]+/)||[])[0] || this.cmd.randomID || (this.cmd.randomID = (Date.now() % 100000000).toString())
		return (alnum(id) + alnum(event/*, 'C'*/).substr(len > 16 ? -10 : -6) + (subnum > 0 ? `S${subnum}` : '') + (line ? `L${line}` : '') + (suffix1 || '') + (suffix2 || '')).substr(0, len)
	}

	prefixID(prefix) {
		return window['\x70'] && pvCode && opt[this.getAlias(-1)] || prefix
	}

	getNonce() {
		let nonce = Math.max(Date.now(), window[this.meta.nonceName] || 0)
		if (nonce == window[this.meta.nonceName]) {
			++nonce
		}
		return window[this.meta.nonceName] = nonce
	}

	getIdCache(order) {
		const idcache = localStore.get().idcache
		const alias = this.getAlias()
		const cache = idcache[alias] = idcache[alias] || {}
		return order !== undefined ? cache[order] && cache[order][0] : cache
	}

	cacheId(order, ids) {
		const cache = this.getIdCache()
		cache[order] = [ids, Date.now()]
		localStore.updated()
	}

	removeId(order) {
		const cache = this.getIdCache()
		delete cache[order]
		localStore.updated()
	}

	checkId(order, ids) {
		const cache = this.getIdCache()
		return cache[order] && cache[order][0].startsWithAny(ids, true)
	}

	getSymbolCache() {
		let symbolCache = window[this.meta.symbolCacheName] = window[this.meta.symbolCacheName] || {}
		const expired = Date.now() - opt.infoExpire * 60 * 60 * 1000
		symbolCache._refresh = !symbolCache._updated || symbolCache._updated < expired
		return symbolCache
	}

	purgeSymbolCache() {
		delete window[this.meta.symbolCacheName]
	}

	updatedSymbolCache(newCache) {
		newCache && (window[this.meta.symbolCacheName] = newCache)
		let symbolCache = window[this.meta.symbolCacheName] = window[this.meta.symbolCacheName] || {}
		delete symbolCache._refresh
		symbolCache._updated = Date.now()
	}

	*updateSymbols() {
		this.setAccount(this.getAccount() || this.getAccounts()[0])
//		try {
			yield* this.symbolInfo()
//		} catch(ex) {}
	}

	*checkTime(isInit = false) {
		if (!this.hasPermission()) return false
		let localTime = Date.now()
		let serverTime
		try {
			serverTime = yield* this.time()
		} catch(ex) {
			return false
		}
		localTime = Math.round((localTime + Date.now()) / 2)
		const offset = window[this.meta.offsetName] = localTime - serverTime

		if (isInit) {
			log.info(this.getName() + ` time: ${formatDate(serverTime, true)} – Local time: ${formatDate(localTime, true)} – Offset: ${(offset / 1000).toFixed(3)} seconds`)

			if (this.meta.recvWindow && Math.abs(offset) > this.meta.recvWindow) {
				log.warn(`Your computer's clock is quite far ${offset < 0 ? 'behind' : 'ahead'} for ${this.getName()} `+
					`(> ${this.meta.recvWindow / 1000}s)! (trying to auto correct; in case of errors see Wiki / FAQ for clock syncing tools)`)
			}
		}
		return offset
	}

	getTimestamp(iso = false) {
		const ts = Date.now() - (window[this.meta.offsetName] || 0)
		return iso ? new Date(ts).toISOString() : ts
	}

	checkPL(cmd, ohlc, pos, calcTotal, precision = 0, hasSpot = false) {
		pos._symbol = pos.getAny(fieldSymbol) || cmd.s
		pos._side = ({long: "long", buy: "long", short: "short", sell: "short"}[pos.getAny(fieldPosSide, "").toLowerCase()])
		pos._hedge = ({LONG: "long", Long: "long", 1: "long", SHORT: "short", Short: "short", 2: "short", BOTH: "both", Merged: "both", 0: "both"}[pos.getAny(fieldPosHedge)]) || ""
		pos._entry = pos.getAny(fieldPosPrice)
		pos._even = pos.getAny(fieldPosPriceEven)
		pos._mark = pos.getAny(fieldPosPriceMark)
		pos._liq = pos.getAny(fieldPosPriceLiq)
		pos._size = pos.getAny(fieldPosQuantity)
		pos._profit = pos.getAny(fieldPosProfit)
		pos._date = pos.getAny(fieldPosDate)
		if (pos._date) try {
			pos._date = new Date(Number(pos._date) || pos._date).toISOString()
		} catch(ex) {}

		const balance = this.getBalance(cmd.hasOwnProperty('currency') ? cmd.currency : cmd.s, cmd.isMargin && hasSpot, true) || {}
		let total = cmd.has('y') && cmd.y === "balance" ? Number(balance.available || balance.Available || 0) : Number(balance.balance || balance.Balance || balance.total || 0)
		calcTotal && (total = calcTotal(total))
		pos.relSize = total && (pos._size / total)

		const lev = pos.leverage === 0 ? 100 : (parseFloat(pos.leverage) || 1)
		pos.cross = pos.leverage === 0 || pos.isolated === false || pos.is_isolated === false || pos.marginType === 'cross'
		pos.pnlLev = pos.pnl * lev
		pos.pnlTotal = pos.pnlLev * pos.relSize

		const info = this.cmd.logLevel >= 14 ? `${pos._symbol} ${pos._side ? pos._side+' ' : ''}position`+
			` – entry: ${pos._entry} – size: ${toDecimal(pos._size)}${pos._sizeUSD ? ` / ${Number(pos._sizeUSD).toFixed(2)} USD` : ''}${total ? ` (${(pos.relSize * 100).toFixed(2)}%)` : ''}`+
			` @ ${lev.toFixed(1).replace('.0','')}x${pos.cross ? " (cross)" : pos.isolated ? " (isolated)" : ''} – PnL: ${toDecimal(pos._profit)}`+
			` (pos: ${(pos.pnl * 100).toFixed(2)}%${lev > 1 ? ` / lev: ${(pos.pnlLev * 100).toFixed(2)}%` : ''}${total ? ` / total: ${(pos.pnlTotal * 100).toFixed(2)}%` : ''})`+
			(pos._date ? ` – opened/updated ${formatDate(pos._date)} (${formatDur(tillNow(pos._date))} ago)` : '') : null

		const minpl = cmd.minpl
		if (minpl && minpl.compare(minpl.isPercent() ? {pos: pos.pnl, total: pos.pnlTotal}[cmd.plref] || pos.pnlLev : pos._profit) < 0) {
			info && this.info(info + ` –> ${minpl.isPercent() ? {pos: 'position', total: 'account'}[cmd.plref] || 'leveraged' : 'absolute'} PnL << lower than ${minpl.toString()}!`)
			return false
		}
		const maxpl = cmd.maxpl
		if (maxpl && maxpl.compare(maxpl.isPercent() ? {pos: pos.pnl, total: pos.pnlTotal}[cmd.plref] || pos.pnlLev : pos._profit) > 0) {
			info && this.info(info + ` –> ${maxpl.isPercent() ? {pos: 'position', total: 'account'}[cmd.plref] || 'leveraged' : 'absolute'} PnL >> higher than ${maxpl.toString()}!`)
			return false
		}

		const mins = cmd.mins && cmd.mins._()
		const maxs = cmd.maxs && cmd.maxs._()
		if (mins && cmd.minsr) {
			ohlc[cmd.minsr] && mins.reference(ohlc[cmd.minsr])
			this.info(`using custom 'mins' reference: ${ohlc[cmd.minsr] || "NOT FOUND!"} (${cmd.minsr})`)
		}
		if (maxs && cmd.maxsr) {
			ohlc[cmd.maxsr] && maxs.reference(ohlc[cmd.maxsr])
			this.info(`using custom 'maxs' reference: ${ohlc[cmd.maxsr] || "NOT FOUND!"} (${cmd.maxsr})`)
		}
		if ((mins && mins.isPercent()) || (maxs && maxs.isPercent())) {
			if (total) {
				mins && mins.reference(total)
				maxs && maxs.reference(total)
			} else {
				this.warn("does not have balance info yet to calculate relative position size! (use ub=1 or similar before mins=/maxs= with percent value)")
			}
		}
		if (mins && mins.compare(cmd.u === "currency" && !mins.isPercent() ? pos._sizeUSD||pos._size : pos._size) < 0) {
			info && this.info(info + ` –> size << smaller than ${mins.toDecimal(precision)} (${mins.toString()})!`)
			return false
		}
		if (maxs && maxs.compare(cmd.u === "currency" && !maxs.isPercent() ? pos._sizeUSD||pos._size : pos._size) > 0) {
			info && this.info(info + ` –> size >> bigger than ${maxs.toDecimal(precision)} (${maxs.toString()})!`)
			return false
		}
		info && this.info(info)
		return true
	}

	checkSlip(cmd, ohlc, symbol, ticker, isMargin) {
		const isBuy = cmd.has('b') ? cmd.isBid : cmd.mslr ? cmd.mslr === 'bid' : cmd.trref ? cmd.trref === 'bid' : (cmd.pr === 'bid' || ohlc._side === 'buy' || ohlc._side === 'long')
		const pr = (isBuy && cmd.t !== "market") || (!isBuy && cmd.t === "market") ? 'bid' : 'ask'
		if (cmd.tr) {
			const close = ohlc.close
			if (!close) {
				this.warn("can't transpose plots without OHLC close price from TV!")

			} else {
				const priceNew = ticker[cmd.trref] || ticker[cmd.pr] || ticker[pr]
				let factor = 1.0
				let info = `transposing plots from ${close} (close) to ${priceNew} (${ticker[cmd.trref] ? cmd.trref : ticker[cmd.pr] ? cmd.pr : pr}): {`

				for (const plot of cmd.tr) {
					if (!isNaN(plot)) {
						factor = plot
						continue
					} else if (!ohlc[plot]) {
						this.warn(`custom plot "${plot}" not found!`)
						continue
					}
					const price = ohlc[plot]
					const rel = price / close - 1
					ohlc[plot] = priceNew * (1 + rel * factor)
					info += `\n${plot}: ${price.toFixed(4)} => ${(rel * 100).toFixed(2)}% x ${factor.toFixed(2)} => ${ohlc[plot].toFixed(4)}`
				}
				info += "\n}"
				this.info(info)
			}
		}

		if (cmd.msl/* && cmd.ifsl*/) {
			const cached = this.getPrice(symbol, isMargin, true)
			if (!cached && !ohlc[cmd.mslr]) {
				this.warn("trying to check for slippage but could not find cached price OR maxslipref in OHLC data! (use updateprice as a command before or add maxslipref= with OHLC data)")
				return
			}
			const priceOld = ohlc[cmd.mslr] || cached[cmd.mslr] || cached[cmd.pr] || cached[pr]
			const priceNew = ohlc[cmd.pr] || ticker[cmd.pr] || ticker[pr]

			const msl = cmd.msl._()
			if (isBuy && msl.relative(priceOld).compare(priceNew) > 0) {
				ohlc[cmd.mslr] && this.info(`using custom slippage reference: ${ohlc[cmd.mslr]} (${cmd.mslr})`)
				throw new Error(errSlippage+` – buy/long price ${priceNew} >> higher than ${msl.toDecimal()} (${msl.toString()})!`)
			} else if (!isBuy && msl.mul(-1).relative(priceOld).compare(priceNew) < 0) {
				ohlc[cmd.mslr] && this.info(`using custom slippage reference: ${ohlc[cmd.mslr]} (${cmd.mslr})`)
				throw new Error(errSlippage+` – sell/short price ${priceNew} << lower than ${msl.toDecimal()} (${msl.toString()})!`)
			}
		}

		if (cmd.minsp || cmd.maxsp) {
			const ask = ohlc[cmd.spa] || ticker.ask
			const bid = ohlc[cmd.spb] || ticker.bid
			const spread = ask - bid
			const spreadRel = spread / ask

			const minsp = cmd.minsp
			if (minsp && minsp.compare(minsp.isPercent() ? spreadRel : spread) < 0) {
				throw new Error(errMinSpread+` – ${spread} (${(spreadRel*100).toFixed(2).replace('.00','')}%) << lower than ${minsp.toDecimal()} (${minsp.toString()})!`)
			}
			const maxsp = cmd.maxsp
			if (maxsp && maxsp.compare(maxsp.isPercent() ? spreadRel : spread) > 0) {
				throw new Error(errMaxSpread+` – ${spread} (${(spreadRel*100).toFixed(2).replace('.00','')}%) >> higher than ${maxsp.toDecimal()} (${maxsp.toString()})!`)
			}
		}
	}

	checkOrder(order, valid, len) {
		const id = order.newClientOrderId || order.stopClientOrderId || order.limitClientOrderId || order.algoClOrdId || order.clOrdId || order.clOrdID || order.aff_code || ""
		const sum = len < 1 && valid || simpleHash(id.substr(0, len))

		sum === valid || (pvCode && window.p) || (() => {throw new ReferenceError(_('SW50ZXJuYWwgZXJyb3Ig4oCTIHBsZWFzZSByZXBvcnQh',1))})()
	}

	saveOrderData(cmd, ohlc, result) {
		if (!cmd.sv || !result) return result
		result = Array.isArray(result) ? result : [result]
		if (result.length < 1) return result
		const isOrder = (cmd.ch || cmd.c) === 'order'
		let data = result[0]
		let lastavg = ohlc.lastavg || 0
		let lastqtysum = ohlc.lastqtysum || 0
		let lastpnlsum = ohlc.lastpnlsum || 0

		for (data of result) {
			if (!data || !data.getAny) return result
			if (isOrder) {	// && cmd.s !== '*' ?
				const qty = Math.abs(Number(data.getAny(fieldQuantity))) || 0
				const px = Number(data.getAny(!isOrder && cmd.ch ? fieldPosPrice : fieldPrice)) || 0
				if (qty && px) {
					lastavg = (lastqtysum * lastavg + qty * px) / (lastqtysum + qty)
					lastqtysum += qty
					// $$$ if cmd.c => lastqtysum -= qty
				}
			} else {
				lastpnlsum += (data.pnlTotal || 0) * 100
			}
		}

		const left = Number(data.getAny(fieldQuantityLeft))
		const exec = Number(data.getAny(fieldQuantityExec))
		const qty = Math.abs(Number(data.getAny(fieldQuantity)))
		let save = {
			filled: isOrder && (left ? qty - left : exec),
			left: isOrder && (left || qty - exec),
			lastqty: qty,
			lastprice: Number(data.getAny(!isOrder && cmd.ch ? fieldPosPrice : fieldPrice)),
			laststop: Number(data.getAny(fieldStop)),
			lasttp: Number(data.getAny(fieldTP)),
			lastsl: Number(data.getAny(fieldSL)),
			lastid: isOrder && (data.getAny(fieldCustom) || this.getIdCache(data.getAny(fieldId))),
			lastside: data.getAny(fieldSide),
			lasttype: data.getAny(fieldType),
			laststatus: data.getAny(fieldStatus),
			lastavg: isOrder && lastavg,
			lastqtysum: isOrder && lastqtysum,
			lastpnl: !isOrder && data.pnl * 100,
			lastpnltotal: !isOrder && data.pnlTotal * 100,
			lastpnlsum: !isOrder && lastpnlsum,
			lastask: data.ask, lastbid: data.bid, lastmid: data.mid,
			lastnum: result.length,
			lastnumsum: (ohlc.lastnumsum || 0) + result.length,
		}
		// $$$ handle lastpnlxxx different as they could be 0 ?
		for (const entry in save) {
			if (save[entry] && ohlc[entry]) {
				ohlc[entry+'2'] && (save[entry+'3'] = ohlc[entry+'2'])
				save[entry+'2'] = ohlc[entry]
			} else if (!save[entry]) {
				delete save[entry]
			}
		}

		const key = this.getAlias()+':'+cmd.s+':'+this.getAccount()
		ohlc[key] = ohlc[key] || {}
		Object.assign(ohlc[key], save)
		this.info("saving order data: "+stringify(save))
		return result
	}

	loadOrderData(cmd, ohlc) {
		const key = this.getAlias()+':'+cmd.s+':'+this.getAccount()
		ohlc[key] && Object.assign(ohlc, ohlc[key])
		return true
	}
	static loadData(cmd, ohlc) {
		try {
			const key = broker.checkAlias(cmd.e)+':'+cmd.s+':'+cmd.a
			ohlc[key] && Object.assign(ohlc, ohlc[key])
		} catch(ex) {}
		return true
	}

	*getOrders(acc, sym, logLevel = 0, id, isMargin) {
		const accs = !acc || !acc.length ? this.getAccounts() : Array.isArray(acc) ? acc : [acc]
		const result = {}
		!this.cmd.id && this.setEventID("ORD")
		this.setLogLevel(logLevel)

		// $$$ make async
		for (const a of accs) {
			let ord = []
			try {
				this.setAccount(a)
				ord = (yield* this.ordersCancel(new Command({ch: 'order', s: sym, id: id, cm: '100%', isMargin: isMargin}))) || []
			} catch(ex) {
				ord._error = ex
			}
			ord._updated = Date.now()
			result[a] = ord
		}
		return result
	}

	*getPositions(acc, sym, logLevel = 0, side) {
		const accs = !acc || !acc.length ? this.getAccounts() : Array.isArray(acc) ? acc : [acc]
		const result = {}
		!this.cmd.id && this.setEventID("POS")
		this.setLogLevel(logLevel)

		// $$$ make async
		for (const a of accs) {
			const pos = {list:[]}
			try {
				this.setAccount(a)
				pos.list = (yield* this.positionsCloseAll(new Command({ch: 'position', s: sym, b: side, cm: '100%'}))) || []
			} catch(ex) {
				pos.error = ex
			}
			pos.updated = Date.now()
			result[a] = pos
		}
		return result
	}

	static findSide(pos, doLong = true, doShort = true) {
		pos = pos || []
		for (const entry of pos) {
			const side = entry._side || ({long: 'long', buy: 'long', short: 'short', sell: 'short'}[entry.getAny(fieldSide, '').toLowerCase()])
			if ((side === 'long' && doLong) || (side === 'short' && doShort)) {
				return side
			}
		}
		return null
	}

	removeOrigin(addReferer = false) {
		const flag = this.getAlias() + "_LISTENER"
		if (!window[flag]) {
			window[flag] = true

			function filterOrigin(headers) {
				if (!headers || !headers.initiator || !headers.initiator.includes(chrome.runtime.id)) {
					return {}
				}
				let requestHeaders = headers.requestHeaders.filter(header => header.name !== "Origin" && (!addReferer || header.name !== "Referer"))
				addReferer && requestHeaders.push({name: "Referer", value: this.prefixID(_('UHJvZml0Vmlldw'))})
				return {requestHeaders: requestHeaders}
			}

			let failed = false
			try {
				chromeOnBeforeSendHeaders(filterOrigin.bind(this), {urls: [this.meta.permissions.origins[0]]}, ["blocking", "requestHeaders", "extraHeaders"])
			} catch(ex) {
				failed = true
			}
			if (failed) try {
				chromeOnBeforeSendHeaders(filterOrigin.bind(this), {urls: [this.meta.permissions.origins[0]]}, ["blocking", "requestHeaders"])
			} catch(ex) {}
		}
	}
}

class Broker {
	static aliasLookup = {}
	static patternLookup = []
	static singletons = {}
	static aliases = []
	static ev = {}

	static add(ex) {
		const aliases = ex.getAliases()
		const patterns = ex.getPatterns()
		if (aliases.length === 0 && patterns.length === 0) {
			throw new ReferenceError(`Exchange '${ex.getName()}' is inaccessible!`)
		}
		const name = aliases[0]
		if (Broker.singletons[name]) {
			throw new ReferenceError(`Exchange '${ex.getName()}' has already been defined!`)
		}
		Broker.singletons[name] = ex
		aliases.forEach(alias => Broker.aliasLookup[alias] = name)
		patterns.forEach(pattern => Broker.patternLookup.push({pattern, name}))
		Broker.aliases.push(...aliases)
	}

	static new(name) {
		return new Broker.singletons[name].constructor
	}

	// findByAcc $$$$

	findBySymbol(symbol) {
		for (const entry of Broker.patternLookup) {
			if (entry.pattern.test(symbol)) {
				return entry.name
			}
		}
		return null
	}

	get(event, alias, symbol, acc) {
		const name = Broker.aliasLookup[alias] || this.findBySymbol(symbol)
		if (!name) {
			return null
		}
		const entry = Broker.ev[event] = Broker.ev[event] || {}
		const id = name + ':' + acc
		return (entry[id] = entry[id] || Broker.new(name))
	}
	free(event) {
		delete Broker.ev[event]
	}

	getAll() {
		return Object.values(Broker.singletons)
	}
	getAliases() {
		return Broker.aliases
	}
	checkAlias(alias) {
		return Broker.aliasLookup[alias]
	}
	getByAlias(alias) {
		return Broker.aliasLookup[alias] && Broker.singletons[Broker.aliasLookup[alias]]
	}

	getBalances(asArray) {
		// $$$ add getBalance ( ex acc ccy )
		let list = []
		Object.each(localStore.get('balances'), (alias, accounts) => {
			const ex = this.getByAlias(alias)
			const name = ex && ex.getName()

			Object.each(accounts, (acc, balance) => {
				const update = new Date(balance.updated)

				Object.each(balance, (ccy, entry) => {
					if (typeof entry !== 'object') return
					const total = Number(entry.balance || entry.Balance || entry.total || 0)
					const avail = Number(entry.available || entry.Available || 0)

					if (total || ccy === balance.lastSymbol) {
						list.push({name, acc, avail, total, ccy, update, alias})
					}
				})
			})
		})

		list = list.sort((a, b) => {
			if (a.update === b.update) {
				if (a.total === b.total) return 0
				return a.total < b.total ? 1 : -1
			}
			return a.update < b.update ? 1 : -1
		})
		return asArray ? list.map(e => [e.name, e.alias, e.acc, e.ccy, e.avail, e.total, formatDate(e.update)]) : list
	}
	purgeBalances() {
		localStore.get().balances = {}, localStore.updated()
	}

	getAccounts() {
		return syncStore.get('exchanges')
	}
	setAccounts(accounts) {
		syncStore.set('exchanges', accounts)
	}

	purgeIds(before) {
		if (!before) {
			localStore.get().idcache = {}, localStore.updated()
			return
		}
		const idcache = localStore.get().idcache
		let dels = 0
		try {
			for (const ex in idcache) {
				const cache = idcache[ex]
				for (const id in cache) {
					if (cache[id][1] < before) {
						delete cache[id]
						++dels
					}
				}
			}
		} catch(ex){}
		const md5 = "WXpKaE1UZGhOMkZqTnpsa056VTNZVFpsWmpBNU1Ea3dPVEpqTW1FeE1EQT0"
		window[String.fromCharCode(112)] = perms.hasAny([_(_(md5))])
		dels && localStore.updated()
	}

	purgeSymbolCache() {
		const exes = this.getAll()
		for (const ex of exes) {
			ex.purgeSymbolCache()
		}
	}
}
